from __future__ import annotations


class TomlCombineError(Exception):
    """There was an error in the toml-combine library."""

    def __init__(self, message: str = "", **kwargs) -> None:
        message = message or type(self).__doc__ or ""
        message = message.format(**kwargs)

        super().__init__(message)


class TomlDecodeError(TomlCombineError):
    """Error while decoding configuration file."""


class TomlEncodeError(TomlCombineError):
    """Error while encoding configuration file."""


class IncompatibleOverrides(TomlCombineError):
    """Incompatible overrides `{id}` and `{other_override}`: When they're both applicable, overrides defining a common overridden key ({key}) must be a subset of one another"""


class DimensionNotFound(TomlCombineError):
    """In {type} {id}: Dimension {dimension} not found."""


class DimensionValueNotFound(TomlCombineError):
    """In {type} {id}: Value {value} for dimension {dimension} not found."""


class MissingOverrideCondition(TomlCombineError):
    """In override {id}: Missing 'when' key in override configuration"""



from __future__ import annotations

import pytest

from toml_combine import combiner, exceptions, toml


@pytest.mark.parametrize(
    "a, b, expected",
    [
        pytest.param(
            {"a": 1, "b": 2},
            {"b": 3, "c": 4},
            {"a": 1, "b": 3, "c": 4},
            id="normal_dicts",
        ),
        pytest.param(
            {"a": {"b": 1, "c": 2}},
            {"a": {"c": 3}},
            {"a": {"b": 1, "c": 3}},
            id="nested_dicts",
        ),
    ],
)
def test_merge_configs__dicts(a, b, expected):
    assert combiner.merge_configs(a, b) == expected


def test_merge_configs__dicts_error():
    with pytest.raises(ValueError):
        combiner.merge_configs({"a": 1}, {"a": {"b": 2}})


@pytest.mark.parametrize(
    "mapping, override, expected",
    [
        (
            {"env": "dev"},
            combiner.Override(when={"env": ["dev"]}, config={}),
            True,
        ),
        (
            {"env": "dev"},
            combiner.Override(when={"env": ["staging"]}, config={}),
            False,
        ),
        (
            {"env": "dev"},
            combiner.Override(when={"env": ["dev", "staging"]}, config={}),
            True,
        ),
        (
            {"env": "staging"},
            combiner.Override(when={"region": ["us"]}, config={}),
            False,
        ),
        (
            {"env": "dev"},
            combiner.Override(when={"env": ["dev"], "region": ["us"]}, config={}),
            False,
        ),
        (
            {"env": "dev", "region": "us"},
            combiner.Override(when={"env": ["dev"]}, config={}),
            True,
        ),
    ],
)
def test_mapping_matches_override(mapping, override, expected):
    result = combiner.mapping_matches_override(mapping=mapping, override=override)
    assert result == expected


def test_build_config():
    raw_config = """
    [dimensions]
    env = ["dev", "staging", "prod"]
    region = ["eu"]

    [default]
    foo = "bar"

    [[override]]
    when.env = ["dev", "staging"]
    when.region = ["eu"]
    foo = "baz"

    [[override]]
    when.env = "dev"
    foo = "qux"
    """

    config_dict = toml.loads(raw_config)
    config = combiner.build_config(config_dict)

    assert config == combiner.Config(
        dimensions={"env": ["dev", "staging", "prod"], "region": ["eu"]},
        default={"foo": "bar"},
        overrides=[
            # Note: The order of the overrides is important: more specific overrides
            # must be listed last.
            combiner.Override(
                when={"env": ["dev"]},
                config={"foo": "qux"},
            ),
            combiner.Override(
                when={"env": ["dev", "staging"], "region": ["eu"]},
                config={"foo": "baz"},
            ),
        ],
    )


def test_build_config__dimension_not_found_in_override():
    raw_config = """
    [dimensions]
    env = ["dev", "prod"]

    [[override]]
    when.region = "eu"
    """

    config = toml.loads(raw_config)
    with pytest.raises(exceptions.DimensionNotFound):
        combiner.build_config(config)


def test_build_config__dimension_value_not_found_in_override():
    raw_config = """
    [dimensions]
    env = ["dev", "prod"]

    [[override]]
    when.env = "staging"
    """

    config = toml.loads(raw_config)
    with pytest.raises(exceptions.DimensionValueNotFound):
        combiner.build_config(config)


def test_extract_keys():
    config = toml.loads(
        """
        a = 1
        b.c = 1
        b.d = 1
        e.f.g = 1
        """,
    )

    result = list(combiner.extract_keys(config))
    assert result == [
        ("a",),
        ("b", "c"),
        ("b", "d"),
        ("e", "f", "g"),
    ]


@pytest.mark.parametrize(
    "a, b, expected",
    [
        pytest.param(
            {"env": ["dev"], "region": ["eu"]},
            {"env": ["dev"]},
            True,
            id="subset1",
        ),
        pytest.param(
            {"env": ["dev"]},
            {"env": ["dev"], "region": ["eu"]},
            True,
            id="subset2",
        ),
        pytest.param(
            {"env": ["prod"], "region": ["eu"]},
            {"env": ["dev"]},
            True,
            id="subset3",
        ),
        pytest.param(
            {"env": ["dev"]},
            {"env": ["prod"], "region": ["eu"]},
            True,
            id="subset4",
        ),
        pytest.param({"env": ["dev"]}, {"region": ["eu"]}, False, id="disjoint"),
        pytest.param(
            {"env": ["dev"], "service": ["frontend"]},
            {"region": ["eu"], "service": ["frontend"]},
            False,
            id="overlap",
        ),
        pytest.param({"env": ["dev"]}, {"env": ["dev"]}, False, id="same_keys1"),
        pytest.param(
            {"env": ["dev", "prod"]}, {"env": ["dev"]}, False, id="same_keys1"
        ),
        pytest.param(
            {"env": ["prod"]}, {"env": ["dev"]}, True, id="same_keys_disjoint"
        ),
        pytest.param(
            {"env": ["prod", "staging"]},
            {"env": ["dev", "sandbox"]},
            True,
            id="multiple_keys_disjoint",
        ),
    ],
)
def test_are_conditions_compatible(a, b, expected):
    assert combiner.are_conditions_compatible(a, b) == expected


@pytest.mark.parametrize(
    "mapping, expected",
    [
        (
            {"env": "prod"},
            {"foo": "bar"},
        ),
        (
            {"env": "staging"},
            {"foo": "baz"},
        ),
    ],
)
def test_generate_for_mapping__simple_case(mapping, expected):
    config = combiner.build_config(
        toml.loads(
            """
            [dimensions]
            env = ["prod", "staging"]

            [default]
            foo = "bar"

            [[override]]
            when.env = "staging"
            foo = "baz"
            """,
        )
    )
    result = combiner.generate_for_mapping(
        config=config,
        mapping=mapping,
    )
    assert result == expected


@pytest.mark.parametrize(
    "mapping, expected",
    [
        pytest.param(
            {"env": "dev"},
            {
                "a": 1,
                "b": 2,
                "c": 3,
                "d": {"e": {"h": {"i": {"j": 4}}}},
                "g": 6,
            },
            id="no_matches",
        ),
        pytest.param(
            {"env": "prod"},
            {
                "a": 10,
                "b": 2,
                "c": 30,
                "d": {"e": {"h": {"i": {"j": 40}}}},
                "g": 60,
            },
            id="single_match",
        ),
        pytest.param(
            {"env": "staging"},
            {
                "a": 1,
                "b": 200,
                "c": 300,
                "d": {"e": {"h": {"i": {"j": 400}}}},
                "f": 500,
                "g": 6,
            },
            id="dont_override_if_match_is_more_specific",
        ),
    ],
)
def test_generate_for_mapping__complex_case(mapping: dict, expected: dict[str, int]):
    default = {
        "a": 1,
        "b": 2,
        "c": 3,
        "d": {"e": {"h": {"i": {"j": 4}}}},
        "g": 6,
    }

    overrides = [
        combiner.Override(
            when={"env": ["prod"]},
            config={
                "a": 10,
                "c": 30,
                "d": {"e": {"h": {"i": {"j": 40}}}},
                "g": 60,
            },
        ),
        combiner.Override(
            when={"env": ["staging"]},
            config={
                "b": 200,
                "c": 300,
                "d": {"e": {"h": {"i": {"j": 400}}}},
                "f": 500,
            },
        ),
        combiner.Override(
            when={"env": ["staging"], "region": ["us"]},
            config={"f": 5000, "g": 6000},
        ),
    ]
    config = combiner.Config(
        dimensions={"env": ["prod", "staging"], "region": ["us"]},
        default=default,
        overrides=overrides,
    )

    result = combiner.generate_for_mapping(config=config, mapping=mapping)
    assert result == expected


def test_generate_for_mapping__duplicate_overrides():
    raw_config = """
    [dimensions]
    env = ["prod"]

    [[override]]
    when.env = "prod"
    foo = "baz"

    [[override]]
    when.env = "prod"
    foo = "qux"
    """

    config = combiner.build_config(toml.loads(raw_config))
    with pytest.raises(exceptions.IncompatibleOverrides):
        combiner.generate_for_mapping(config=config, mapping={"env": "prod"})


def test_generate_for_mapping__duplicate_overrides_different_vars():
    raw_config = """
    [dimensions]
    env = ["prod"]

    [[override]]
    when.env = "prod"
    foo = "baz"

    [[override]]
    when.env = "prod"
    baz = "qux"
    """

    config = combiner.build_config(toml.loads(raw_config))
    assert combiner.generate_for_mapping(config=config, mapping={"env": "prod"}) == {
        "foo": "baz",
        "baz": "qux",
    }


def test_generate_for_mapping__duplicate_overrides_list():
    raw_config = """
    [dimensions]
    env = ["prod", "dev"]

    [[override]]
    when.env = ["prod"]
    hello.world = 1

    [[override]]
    when.env = ["prod", "dev"]
    hello.world = 2
    """

    config = combiner.build_config(toml.loads(raw_config))
    with pytest.raises(exceptions.IncompatibleOverrides) as excinfo:
        combiner.generate_for_mapping(config=config, mapping={"env": "prod"})

    # Message is a bit complex so we test it too.
    assert (
        str(excinfo.value)
        == "Incompatible overrides `{'env': ['prod', 'dev']}` and `{'env': ['prod']}`: "
        "When they're both applicable, overrides defining a common overridden key (hello.world) "
        "must be a subset of one another"
    )



from __future__ import annotations

import json
import pathlib

from toml_combine import cli, toml


def test_cli__json(capsys):
    """Test the CLI."""
    exit_code = cli.cli(
        argv=[
            "tests/test.toml",
            "--format",
            "json",
            "--environment",
            "staging",
            "--type",
            "service",
            "--stack",
            "django",
            "--service",
            "admin",
        ]
    )
    out, err = capsys.readouterr()
    print("out:")
    print(out)
    print("err:")
    print(err)
    assert exit_code == 0

    expected = json.loads((pathlib.Path(__file__).parent / "result.json").read_text())
    assert json.loads(out) == expected["staging-service-django-admin"]


def test_cli__toml(capsys):
    """Test the CLI."""
    cli.cli(
        argv=[
            "tests/test.toml",
            "--environment",
            "staging",
            "--type",
            "service",
            "--stack",
            "django",
            "--service",
            "admin",
        ]
    )
    out, err = capsys.readouterr()
    print("out:")
    print(out)
    print("err:")
    print(err)

    expected = json.loads((pathlib.Path(__file__).parent / "result.json").read_text())
    assert toml.loads(out) == expected["staging-service-django-admin"]



from __future__ import annotations

import json
import pathlib

import pytest

import toml_combine
from toml_combine import toml

config_file = pathlib.Path(__file__).parent / "test.toml"


@pytest.fixture
def expected():
    return json.loads((pathlib.Path(__file__).parent / "result.json").read_text())


@pytest.mark.parametrize(
    "mapping, expected_key",
    [
        pytest.param(
            {
                "environment": "staging",
                "type": "service",
                "stack": "next",
            },
            "staging-service-next",
            id="staging-service-next",
        ),
        pytest.param(
            {
                "environment": "staging",
                "type": "service",
                "stack": "django",
                "service": "api",
            },
            "staging-service-django-api",
            id="staging-service-django-api",
        ),
        pytest.param(
            {
                "environment": "staging",
                "type": "service",
                "stack": "django",
                "service": "admin",
            },
            "staging-service-django-admin",
            id="staging-service-django-admin",
        ),
        pytest.param(
            {
                "environment": "staging",
                "type": "job",
                "stack": "django",
                "job": "manage",
            },
            "staging-job-django-manage",
            id="staging-job-django-manage",
        ),
        pytest.param(
            {
                "environment": "staging",
                "type": "job",
                "stack": "django",
                "job": "special-command",
            },
            "staging-job-django-special-command",
            id="staging-job-django-special-command",
        ),
        pytest.param(
            {
                "environment": "production",
                "type": "service",
                "stack": "next",
            },
            "production-service-next",
            id="production-service-next",
        ),
        pytest.param(
            {
                "environment": "production",
                "type": "service",
                "stack": "django",
                "service": "api",
            },
            "production-service-django-api",
            id="production-service-django-api",
        ),
        pytest.param(
            {
                "environment": "production",
                "type": "service",
                "stack": "django",
                "service": "admin",
            },
            "production-service-django-admin",
            id="production-service-django-admin",
        ),
        pytest.param(
            {
                "environment": "production",
                "type": "job",
                "stack": "django",
                "job": "manage",
            },
            "production-job-django-manage",
            id="production-job-django-manage",
        ),
        pytest.param(
            {
                "environment": "production",
                "type": "job",
                "stack": "django",
                "job": "special-command",
            },
            "production-job-django-special-command",
            id="production-job-django-special-command",
        ),
    ],
)
def test_full_config(mapping, expected, expected_key):
    result = toml_combine.combine(config_file=config_file, **mapping)
    assert result == expected[expected_key]


@pytest.mark.parametrize(
    "kwargs",
    [
        pytest.param({"config_file": config_file}, id="path"),
        pytest.param({"config_file": str(config_file)}, id="path_str"),
        pytest.param({"config": config_file.read_text()}, id="text"),
        pytest.param({"config": toml.loads(config_file.read_text())}, id="parsed"),
    ],
)
def test_full_load_kwargs(kwargs, expected):
    result = toml_combine.combine(
        **kwargs,
        environment="production",
        type="service",
        stack="django",
        service="api",
    )
    assert result == expected["production-service-django-api"]



from __future__ import annotations

import copy
import dataclasses
from collections.abc import Iterable, Mapping, Sequence
from typing import Any, TypeVar

from . import exceptions


@dataclasses.dataclass()
class Override:
    when: Mapping[str, list[str]]
    config: Mapping[str, Any]

    def __str__(self) -> str:
        return f"Override({self.when})"


@dataclasses.dataclass()
class Config:
    dimensions: Mapping[str, list[str]]
    default: Mapping[str, Any]
    # List of overrides, in order of increasing specificity
    overrides: Sequence[Override]


def clean_dimensions_dict(
    to_sort: Mapping[str, list[str]], clean: dict[str, list[str]], type: str
) -> dict[str, list[str]]:
    """
    Recreate a dictionary of dimension values with the same order as the
    dimensions list.
    Also check that the values are valid.
    """
    result = {}
    if invalid_dimensions := set(to_sort) - set(clean):
        raise exceptions.DimensionNotFound(
            type=type,
            id=to_sort,
            dimension=", ".join(invalid_dimensions),
        )

    # Fix the order of the dimensions
    for dimension, valid_values in clean.items():
        if dimension not in to_sort:
            continue

        original_values = to_sort[dimension]
        if invalid_values := set(original_values) - set(valid_values):
            raise exceptions.DimensionValueNotFound(
                type=type,
                id=to_sort,
                dimension=dimension,
                value=", ".join(invalid_values),
            )
        # Fix the order of the values
        result[dimension] = [e for e in valid_values if e in original_values]

    return result


T = TypeVar("T", dict, list, str, int, float, bool)


def merge_configs(a: T, b: T, /) -> T:
    """
    Recursively merge two configuration dictionaries a and b, with b taking precedence.
    """
    if isinstance(a, dict) != isinstance(b, dict):
        raise ValueError(f"Cannot merge {type(a)} with {type(b)}")

    if not isinstance(a, dict):
        return b

    result = a.copy()
    for key, b_value in b.items():  # type: ignore
        if a_value := a.get(key):
            result[key] = merge_configs(a_value, b_value)
        else:
            result[key] = b_value
    return result


def extract_keys(config: Any) -> Iterable[tuple[str, ...]]:
    """
    Extract the keys from a config.
    """
    if isinstance(config, dict):
        for key, value in config.items():
            for sub_key in extract_keys(value):
                yield (key, *sub_key)
    else:
        yield tuple()


def are_conditions_compatible(
    a: Mapping[str, list[str]], b: Mapping[str, list[str]], /
) -> bool:
    """
    `a` and `b` are dictionaries representing override conditions (`when`). Return
    `True` if the conditions represented by `a` are compatible with `b`. Conditions are
    compatible if one is stricly more specific than the other or if they're mutually
    exclusive.
    """
    # Subset
    if set(a) < set(b) or set(b) < set(a):
        return True

    # Disjoint or overlapping sets
    if set(a) != set(b):
        return False

    # Equal sets: it's only compatible if the values are disjoint
    if any(set(a[key]) & set(b[key]) for key in a.keys()):
        return False
    return True


def build_config(config: dict[str, Any]) -> Config:
    """
    Build a finalized Config object from the given configuration dictionary.
    """
    config = copy.deepcopy(config)
    # Parse dimensions
    dimensions = config.pop("dimensions")

    # Parse template
    default = config.pop("default", {})

    overrides = []
    for override in config.pop("override", []):
        try:
            when = override.pop("when")
        except KeyError:
            raise exceptions.MissingOverrideCondition(id=override)
        when = clean_dimensions_dict(
            to_sort={k: v if isinstance(v, list) else [v] for k, v in when.items()},
            clean=dimensions,
            type="override",
        )

        overrides.append(Override(when=when, config=override))

    # Sort overrides by increasing specificity
    overrides = sorted(overrides, key=lambda override: len(override.when))

    return Config(
        dimensions=dimensions,
        default=default,
        overrides=overrides,
    )


def mapping_matches_override(mapping: Mapping[str, str], override: Override) -> bool:
    """
    Check if the values in the override match the given dimensions.
    """
    for dim, values in override.when.items():
        if dim not in mapping:
            return False

        if mapping[dim] not in values:
            return False

    return True


def generate_for_mapping(
    config: Config,
    mapping: Mapping[str, str],
) -> Mapping[str, Any]:
    """
    Generate a configuration based on the provided mapping of dimension values.
    The mapping should contain only the dimensions defined in the config.
    If a dimension is not defined in the mapping, the default value for that
    dimension will be used.
    """

    result = copy.deepcopy(config.default)
    keys_to_conditions: dict[tuple[str, ...], list[Mapping[str, list[str]]]] = {}
    # Apply each matching override
    for override in config.overrides:
        # Check if all dimension values in the override match

        if mapping_matches_override(mapping=mapping, override=override):
            # Check that all applicableoverrides are compatible
            keys = extract_keys(override.config)

            for key in keys:
                previous_conditions = keys_to_conditions.setdefault(key, [])

                for previous_condition in previous_conditions:
                    if not are_conditions_compatible(previous_condition, override.when):
                        raise exceptions.IncompatibleOverrides(
                            id=override.when,
                            key=".".join(key),
                            other_override=previous_condition,
                        )

                keys_to_conditions[key].append(override.when)

            result = merge_configs(result, override.config)

    return result



from __future__ import annotations

import json
from typing import Any

import tomlkit
import tomlkit.exceptions

from . import exceptions


def loads(raw_config: str) -> dict[str, Any]:
    try:
        return tomlkit.loads(raw_config)
    except tomlkit.exceptions.ParseError as e:
        raise exceptions.TomlDecodeError from e


def dumps(config: dict) -> str:
    # https://github.com/python-poetry/tomlkit/issues/411
    # While we're waiting for a fix, the workaround is to give up "style-preserving"
    # features. The easiest way to turn tomlkit objects into plain dicts and strings
    # is through a json round-trip.
    try:
        return tomlkit.dumps(json.loads(json.dumps(config)), sort_keys=False)
    except tomlkit.exceptions.TOMLKitError as e:
        raise exceptions.TomlEncodeError from e
